@@ -4,7 +4,7 @@
  */
 
 /*
- * Copyright (C) 2011-2015 Oracle Corporation
+ * Copyright (C) 2011-2020 Oracle Corporation
  *
  * This file is part of VirtualBox Open Source Edition (OSE), as
  * available from http://www.virtualbox.org. This file is free software;
@@ -15,58 +15,66 @@
  * hope that it will be useful, but WITHOUT ANY WARRANTY of any kind.
  */
 
-#ifdef VBOX_WITH_PRECOMPILED_HEADERS
-# include <precomp.h>
-#else  /* !VBOX_WITH_PRECOMPILED_HEADERS */
-
 /* Qt includes: */
 #include <QApplication>
+#include <QDrag>
 #include <QKeyEvent>
 #include <QStringList>
 #include <QTimer>
 #include <QUrl>
+#include <QWidget>
 
 /* VirtualBox interface declarations: */
-#ifndef VBOX_WITH_XPCOM
-# include "VirtualBox.h"
-#else /* !VBOX_WITH_XPCOM */
-# include "VirtualBox_XPCOM.h"
-#endif /* VBOX_WITH_XPCOM */
+#include <VBox/com/VirtualBox.h>
 
 /* GUI includes: */
-# include "UIDnDHandler.h"
+#include "UIDnDHandler.h"
 #ifdef VBOX_WITH_DRAG_AND_DROP_GH
-# include "CDnDSource.h"
-# ifdef RT_OS_WINDOWS
-#  include "UIDnDDataObject_win.h"
-#  include "UIDnDDropSource_win.h"
-# endif
-# include "UIDnDMIMEData.h"
+#include "CDnDSource.h"
+#ifdef RT_OS_WINDOWS
+# include "UIDnDDataObject_win.h"
+# include "UIDnDDropSource_win.h"
+#endif
+#include "UIDnDMIMEData.h"
 #endif /* VBOX_WITH_DRAG_AND_DROP_GH */
 #include "UIMessageCenter.h"
 #include "UISession.h"
 
 /* COM includes: */
-# include "CConsole.h"
-# include "CGuest.h"
-# include "CGuestDnDSource.h"
-# include "CGuestDnDTarget.h"
-# include "CSession.h"
-
-#endif /* !VBOX_WITH_PRECOMPILED_HEADERS */
+#include "CConsole.h"
+#include "CGuest.h"
+#include "CGuestDnDSource.h"
+#include "CGuestDnDTarget.h"
+#include "CSession.h"
 
 #ifdef LOG_GROUP
  #undef LOG_GROUP
 #endif
 #define LOG_GROUP LOG_GROUP_GUEST_DND
 #include <VBox/log.h>
+#include <iprt/err.h>
+
+#if 0
+# ifdef DEBUG
+#  include <QTextStream>
+#  include <QFile>
+/** Enable this to log debug output of a Qt debug build to a file defined by DEBUG_DND_QT_LOGFILE. */
+#  define DEBUG_DND_QT
+#  ifdef RT_OS_WINDOWS
+#   define DEBUG_DND_QT_LOGFILE "c:\\temp\\VBoxQt.log"
+#  else
+#   define DEBUG_DND_QT_LOGFILE "/var/tmp/vboxqt.log"
+#  endif
+# endif /* DEBUG */
+#endif
 
 
 UIDnDHandler::UIDnDHandler(UISession *pSession, QWidget *pParent)
     : m_pSession(pSession)
     , m_pParent(pParent)
-    , m_enmMode(DNDMODE_UNKNOWN)
+    , m_enmOpMode(DNDMODE_UNKNOWN)
     , m_fIsPending(false)
+    , m_fDataRetrieved(false)
 #ifndef RT_OS_WINDOWS
     , m_pMIMEData(NULL)
 #endif
@@ -88,11 +96,11 @@ Qt::DropAction UIDnDHandler::dragEnter(ulong screenID, int x, int y,
                                        Qt::DropAction proposedAction, Qt::DropActions possibleActions,
                                        const QMimeData *pMimeData)
 {
-    LogFlowFunc(("enmMode=%RU32, screenID=%RU32, x=%d, y=%d, action=%ld\n",
-                 m_enmMode, screenID, x, y, toVBoxDnDAction(proposedAction)));
+    LogFlowFunc(("enmOpMode=%RU32, screenID=%RU32, x=%d, y=%d, action=%ld\n",
+                 m_enmOpMode, screenID, x, y, toVBoxDnDAction(proposedAction)));
 
-    if (   m_enmMode != DNDMODE_UNKNOWN
-        && m_enmMode != DNDMODE_HOSTTOGUEST)
+    if (   m_enmOpMode != DNDMODE_UNKNOWN
+        && m_enmOpMode != DNDMODE_HOSTTOGUEST)
         return Qt::IgnoreAction;
 
     /* Ask the guest for starting a DnD event. */
@@ -103,20 +111,22 @@ Qt::DropAction UIDnDHandler::dragEnter(ulong screenID, int x, int y,
                                           toVBoxDnDActions(possibleActions),
                                           pMimeData->formats().toVector());
     if (m_dndTarget.isOk())
-        setMode(DNDMODE_HOSTTOGUEST);
+    {
+        setOpMode(DNDMODE_HOSTTOGUEST);
+        return toQtDnDAction(result);
+    }
 
-    /* Set the DnD action returned by the guest. */
-    return toQtDnDAction(result);
+    return Qt::IgnoreAction;
 }
 
 Qt::DropAction UIDnDHandler::dragMove(ulong screenID, int x, int y,
                                       Qt::DropAction proposedAction, Qt::DropActions possibleActions,
                                       const QMimeData *pMimeData)
 {
-    LogFlowFunc(("enmMode=%RU32, screenID=%RU32, x=%d, y=%d, action=%ld\n",
-                 m_enmMode, screenID, x, y, toVBoxDnDAction(proposedAction)));
+    LogFlowFunc(("enmOpMode=%RU32, screenID=%RU32, x=%d, y=%d, action=%ld\n",
+                 m_enmOpMode, screenID, x, y, toVBoxDnDAction(proposedAction)));
 
-    if (m_enmMode != DNDMODE_HOSTTOGUEST)
+    if (m_enmOpMode != DNDMODE_HOSTTOGUEST)
         return Qt::IgnoreAction;
 
     /* Notify the guest that the mouse has been moved while doing
@@ -127,18 +137,20 @@ Qt::DropAction UIDnDHandler::dragMove(ulong screenID, int x, int y,
                                          toVBoxDnDAction(proposedAction),
                                          toVBoxDnDActions(possibleActions),
                                          pMimeData->formats().toVector());
-    /* Set the DnD action returned by the guest. */
-    return toQtDnDAction(result);
+    if (m_dndTarget.isOk())
+        return toQtDnDAction(result);
+
+    return Qt::IgnoreAction;
 }
 
 Qt::DropAction UIDnDHandler::dragDrop(ulong screenID, int x, int y,
                                       Qt::DropAction proposedAction, Qt::DropActions possibleActions,
                                       const QMimeData *pMimeData)
 {
-    LogFlowFunc(("enmMode=%RU32, screenID=%RU32, x=%d, y=%d, action=%ld\n",
-                 m_enmMode, screenID, x, y, toVBoxDnDAction(proposedAction)));
+    LogFlowFunc(("enmOpMode=%RU32, screenID=%RU32, x=%d, y=%d, action=%ld\n",
+                 m_enmOpMode, screenID, x, y, toVBoxDnDAction(proposedAction)));
 
-    if (m_enmMode != DNDMODE_HOSTTOGUEST)
+    if (m_enmOpMode != DNDMODE_HOSTTOGUEST)
         return Qt::IgnoreAction;
 
     /* The format the guest requests. */
@@ -155,7 +167,7 @@ Qt::DropAction UIDnDHandler::dragDrop(ulong screenID, int x, int y,
     if (   m_dndTarget.isOk()
         && enmResult != KDnDAction_Ignore)
     {
-        LogFlowFunc(("strFormat=%s ...\n", strFormat.toAscii().constData()));
+        LogFlowFunc(("strFormat=%s ...\n", strFormat.toUtf8().constData()));
 
         QByteArray arrBytes;
 
@@ -180,14 +192,14 @@ Qt::DropAction UIDnDHandler::dragDrop(ulong screenID, int x, int y,
          **/
         else
         {
-            LogRel3(("DnD: Guest requested a different format '%s'\n", strFormat.toAscii().constData()));
+            LogRel3(("DnD: Guest requested a different format '%s'\n", strFormat.toUtf8().constData()));
             LogRel3(("DnD: The host offered:\n"));
 #if 0
             for (QStringList::iterator itFmt  = pMimeData->formats().begin();
                                        itFmt != pMimeData->formats().end(); itFmt++)
             {
                 QString strTemp = *itFmt;
-                LogRel3(("DnD: \t%s\n", strTemp.toAscii().constData()));
+                LogRel3(("DnD: \t%s\n", strTemp.toUtf8().constData()));
             }
 #endif
             if (pMimeData->hasText())
@@ -211,7 +223,7 @@ Qt::DropAction UIDnDHandler::dragDrop(ulong screenID, int x, int y,
             memcpy(vecData.data(), arrBytes.constData(), arrBytes.size());
 
             /* Send data to the guest. */
-            LogRel3(("DnD: Host is sending %d bytes of data as '%s'\n", vecData.size(), strFormat.toAscii().constData()));
+            LogRel3(("DnD: Host is sending %d bytes of data as '%s'\n", vecData.size(), strFormat.toUtf8().constData()));
             CProgress progress = m_dndTarget.SendData(screenID, strFormat, vecData);
 
             if (m_dndTarget.isOk())
@@ -247,22 +259,51 @@ Qt::DropAction UIDnDHandler::dragDrop(ulong screenID, int x, int y,
      * the end of the current transfer direction. So reset the current
      * mode as well here.
      */
-    setMode(DNDMODE_UNKNOWN);
+    setOpMode(DNDMODE_UNKNOWN);
 
     return toQtDnDAction(enmResult);
 }
 
 void UIDnDHandler::dragLeave(ulong screenID)
 {
-    LogFlowFunc(("enmMode=%RU32, screenID=%RU32\n", m_enmMode, screenID));
+    LogFlowFunc(("enmOpMode=%RU32, screenID=%RU32\n", m_enmOpMode, screenID));
 
-    if (m_enmMode == DNDMODE_HOSTTOGUEST)
+    if (m_enmOpMode == DNDMODE_HOSTTOGUEST)
     {
         m_dndTarget.Leave(screenID);
-        setMode(DNDMODE_UNKNOWN);
+        setOpMode(DNDMODE_UNKNOWN);
     }
 }
 
+#ifdef DEBUG_DND_QT
+QTextStream *g_pStrmLogQt = NULL; /* Output stream for Qt debug logging. */
+
+/* static */
+void UIDnDHandler::debugOutputQt(QtMsgType type, const QMessageLogContext &context, const QString &strMessage)
+{
+    QString strMsg;
+    switch (type)
+    {
+    case QtWarningMsg:
+        strMsg += "[W]";
+        break;
+    case QtCriticalMsg:
+        strMsg += "[C]";
+        break;
+    case QtFatalMsg:
+        strMsg += "[F]";
+        break;
+    case QtDebugMsg:
+    default:
+        strMsg += "[D]";
+        break;
+    }
+
+    if (g_pStrmLogQt)
+        (*g_pStrmLogQt) << strMsg << " " << strMessage << endl;
+}
+#endif /* DEBUG_DND_QT */
+
 /*
  * Source -> Frontend.
  */
@@ -270,6 +311,7 @@ void UIDnDHandler::dragLeave(ulong screenID)
 int UIDnDHandler::dragStartInternal(const QStringList &lstFormats,
                                     Qt::DropAction defAction, Qt::DropActions actions)
 {
+    RT_NOREF(defAction);
     int rc = VINF_SUCCESS;
 
 #ifdef VBOX_WITH_DRAG_AND_DROP_GH
@@ -278,17 +320,27 @@ int UIDnDHandler::dragStartInternal(const QStringList &lstFormats,
     LogFlowFunc(("Number of formats: %d\n", lstFormats.size()));
 # ifdef DEBUG
     for (int i = 0; i < lstFormats.size(); i++)
-        LogFlowFunc(("\tFormat %d: %s\n", i, lstFormats.at(i).toAscii().constData()));
+        LogFlowFunc(("\tFormat %d: %s\n", i, lstFormats.at(i).toUtf8().constData()));
+# endif
+
+# ifdef DEBUG_DND_QT
+    QFile *pFileDebugQt = new QFile(DEBUG_DND_QT_LOGFILE);
+    if (pFileDebugQt->open(QIODevice::WriteOnly | QIODevice::Append | QIODevice::Text))
+    {
+        g_pStrmLogQt = new QTextStream(pFileDebugQt);
+
+        qInstallMessageHandler(UIDnDHandler::debugOutputQt);
+        qDebug("========================================================================");
+    }
 # endif
 
 # ifdef RT_OS_WINDOWS
-
-    UIDnDDropSource *pDropSource = new UIDnDDropSource(m_pParent);
-    if (!pDropSource)
-        return VERR_NO_MEMORY;
     UIDnDDataObject *pDataObject = new UIDnDDataObject(this, lstFormats);
     if (!pDataObject)
         return VERR_NO_MEMORY;
+    UIDnDDropSource *pDropSource = new UIDnDDropSource(m_pParent, pDataObject);
+    if (!pDropSource)
+        return VERR_NO_MEMORY;
 
     DWORD dwOKEffects = DROPEFFECT_NONE;
     if (actions)
@@ -326,13 +378,13 @@ int UIDnDHandler::dragStartInternal(const QStringList &lstFormats,
         return VERR_NO_MEMORY;
     }
 
-    /* Invoke this handler as data needs to be retrieved. */
-    connect(m_pMIMEData, SIGNAL(getData(QString, QVariant::Type, QVariant&)),
-            this, SLOT(sltGetData(QString, QVariant::Type, QVariant&)));
-
     /* Inform the MIME data object of any changes in the current action. */
-    connect(pDrag, SIGNAL(actionChanged(Qt::DropAction)),
-            m_pMIMEData, SLOT(sltDropActionChanged(Qt::DropAction)));
+    connect(pDrag, &QDrag::actionChanged,
+            m_pMIMEData, &UIDnDMIMEData::sltDropActionChanged);
+
+    /* Invoke this handler as data needs to be retrieved by our derived QMimeData class. */
+    connect(m_pMIMEData, &UIDnDMIMEData::sigGetData,
+            this, &UIDnDHandler::sltGetData);
 
     /*
      * Set MIME data object and start the (modal) drag'n drop operation on the host.
@@ -340,7 +392,21 @@ int UIDnDHandler::dragStartInternal(const QStringList &lstFormats,
      */
     pDrag->setMimeData(m_pMIMEData);
     LogFlowFunc(("Executing modal drag'n drop operation ...\n"));
-    Qt::DropAction dropAction = pDrag->exec(actions, defAction);
+
+    Qt::DropAction dropAction;
+#  ifdef RT_OS_DARWIN
+#    ifdef VBOX_WITH_DRAG_AND_DROP_PROMISES
+        dropAction = pDrag->exec(actions, defAction);
+#    else
+        /* Without having VBOX_WITH_DRAG_AND_DROP_PROMISES enabled drag and drop
+         * will not work on OS X! It also requires some handcrafted patches within Qt
+         * (which also needs VBOX_WITH_DRAG_AND_DROP_PROMISES set there). */
+        dropAction = Qt::IgnoreAction;
+        rc = VERR_NOT_SUPPORTED;
+#    endif
+#  else /* !RT_OS_DARWIN */
+    dropAction = pDrag->exec(actions, defAction);
+#  endif /* RT_OS_DARWIN */
     LogRel3(("DnD: Ended with dropAction=%ld\n", UIDnDHandler::toVBoxDnDAction(dropAction)));
 
     /* Note: The UIDnDMimeData object will not be not accessible here anymore,
@@ -351,11 +417,27 @@ int UIDnDHandler::dragStartInternal(const QStringList &lstFormats,
 
 # endif /* !RT_OS_WINDOWS */
 
-#else /* VBOX_WITH_DRAG_AND_DROP_GH */
+    reset();
+
+#ifdef DEBUG_DND_QT
+    if (g_pStrmLogQt)
+    {
+        delete g_pStrmLogQt;
+        g_pStrmLogQt = NULL;
+    }
+
+    if (pFileDebugQt)
+    {
+        pFileDebugQt->close();
+        delete pFileDebugQt;
+    }
+#endif /* DEBUG_DND_QT */
+
+#else /* !VBOX_WITH_DRAG_AND_DROP_GH */
 
     rc = VERR_NOT_SUPPORTED;
 
-#endif
+#endif /* VBOX_WITH_DRAG_AND_DROP_GH */
 
     LogFlowFuncLeaveRC(rc);
     return rc;
@@ -366,13 +448,13 @@ int UIDnDHandler::dragCheckPending(ulong screenID)
     int rc;
 #ifdef VBOX_WITH_DRAG_AND_DROP_GH
 
-    LogFlowFunc(("enmMode=%RU32, fIsPending=%RTbool, screenID=%RU32\n", m_enmMode, m_fIsPending, screenID));
+    LogFlowFunc(("enmOpMode=%RU32, fIsPending=%RTbool, screenID=%RU32\n", m_enmOpMode, m_fIsPending, screenID));
 
     {
         QMutexLocker AutoReadLock(&m_ReadLock);
 
-        if (   m_enmMode != DNDMODE_UNKNOWN
-            && m_enmMode != DNDMODE_GUESTTOHOST) /* Wrong mode set? */
+        if (   m_enmOpMode != DNDMODE_UNKNOWN
+            && m_enmOpMode != DNDMODE_GUESTTOHOST) /* Wrong mode set? */
             return VINF_SUCCESS;
 
         if (m_fIsPending) /* Pending operation is in progress. */
@@ -404,16 +486,16 @@ int UIDnDHandler::dragCheckPending(ulong screenID)
     QVector<QString> vecFormats;
     m_dataSource.defaultAction = m_dndSource.DragIsPending(screenID, vecFormats, m_dataSource.vecActions);
 
-    LogRel3(("DnD: Default action is: 0x%x\n", m_dataSource.defaultAction));
-    LogRel3(("DnD: Number of supported guest actions: %d\n", m_dataSource.vecActions.size()));
+    LogRelMax3(10, ("DnD: Default action is: 0x%x\n", m_dataSource.defaultAction));
+    LogRelMax3(10, ("DnD: Number of supported guest actions: %d\n", m_dataSource.vecActions.size()));
         for (int i = 0; i < m_dataSource.vecActions.size(); i++)
-            LogRel3(("DnD: \tAction %d: 0x%x\n", i, m_dataSource.vecActions.at(i)));
+            LogRelMax3(10, ("DnD: \tAction %d: 0x%x\n", i, m_dataSource.vecActions.at(i)));
 
-    LogRel3(("DnD: Number of supported guest formats: %d\n", vecFormats.size()));
+    LogRelMax3(10, ("DnD: Number of supported guest formats: %d\n", vecFormats.size()));
         for (int i = 0; i < vecFormats.size(); i++)
         {
             const QString &strFmtGuest = vecFormats.at(i);
-            LogRel3(("DnD: \tFormat %d: %s\n", i, strFmtGuest.toAscii().constData()));
+            LogRelMax3(10, ("DnD: \tFormat %d: %s\n", i, strFmtGuest.toUtf8().constData()));
         }
 
     LogFlowFunc(("defaultAction=0x%x, vecFormatsSize=%d, vecActionsSize=%d\n",
@@ -466,7 +548,7 @@ int UIDnDHandler::dragStart(ulong screenID)
         return VERR_INVALID_PARAMETER;
     }
 
-    setMode(DNDMODE_GUESTTOHOST);
+    setOpMode(DNDMODE_GUESTTOHOST);
 
     rc = dragStartInternal(m_dataSource.lstFormats,
                            toQtDnDAction(m_dataSource.defaultAction), toQtDnDActions(m_dataSource.vecActions));
@@ -490,7 +572,7 @@ int UIDnDHandler::dragStop(ulong screenID)
 
     NOREF(screenID);
 
-    m_fIsPending = false;
+    reset();
     rc = VINF_SUCCESS;
 
 #else /* !VBOX_WITH_DRAG_AND_DROP_GH */
@@ -505,36 +587,91 @@ int UIDnDHandler::dragStop(ulong screenID)
     return rc;
 }
 
+void UIDnDHandler::reset(void)
+{
+    LogFlowFuncEnter();
+
+    m_fDataRetrieved = false;
+    m_fIsPending = false;
+
+    setOpMode(DNDMODE_UNKNOWN);
+}
+
+int UIDnDHandler::retrieveData(Qt::DropAction          dropAction,
+                               const QString          &strMIMEType,
+                                     QVector<uint8_t> &vecData)
+{
+    if (!strMIMEType.compare("application/x-qt-mime-type-name", Qt::CaseInsensitive))
+        return VINF_SUCCESS;
+
+    int rc;
+    if (!m_fDataRetrieved)
+    {
+        /*
+         * Retrieve the actual data from the guest.
+         */
+        rc = retrieveDataInternal(dropAction, strMIMEType, m_vecData);
+        if (RT_FAILURE(rc))
+            LogRel3(("DnD: Receiving data failed: %Rrc\n", rc));
+        else
+            m_fDataRetrieved = true;
+    }
+    else /* Data already received, supply cached version. */
+        rc = VINF_SUCCESS;
+
+    if (RT_SUCCESS(rc))
+        vecData = m_vecData;
+
+    return rc;
+}
+
 int UIDnDHandler::retrieveData(      Qt::DropAction  dropAction,
-                               const QString        &strMimeType,
+                               const QString        &strMIMEType,
                                      QVariant::Type  vaType,
                                      QVariant       &vaData)
 {
-    return retrieveDataInternal(dropAction, strMimeType, vaType, vaData);
+    QVector<uint8_t> vecData;
+    int rc = retrieveData(dropAction, strMIMEType, vecData);
+    if (RT_SUCCESS(rc))
+    {
+        /* If no/an invalid variant is set, try to guess the variant type.
+         * This can happen on OS X. */
+        if (vaType == QVariant::Invalid)
+            vaType = UIDnDMIMEData::getVariantType(strMIMEType);
+
+        rc = UIDnDMIMEData::getDataAsVariant(vecData, strMIMEType, vaType, vaData);
+    }
+
+    LogFlowFuncLeaveRC(rc);
+    return rc;
 }
 
-int UIDnDHandler::retrieveDataInternal(      Qt::DropAction  dropAction,
-                                       const QString        &strMimeType,
-                                             QVariant::Type  vaType,
-                                             QVariant       &vaData)
+int UIDnDHandler::retrieveDataInternal(      Qt::DropAction    dropAction,
+                                       const QString          &strMIMEType,
+                                             QVector<uint8_t> &vecData)
 {
-    LogFlowFunc(("Retrieving data as type=%s (variant type=%RU32)\n",
-                 strMimeType.toAscii().constData(), vaType));
+    LogRel3(("DnD: Retrieving data from guest as '%s' (%d)\n", qPrintable(strMIMEType), dropAction));
 
     int rc = VINF_SUCCESS;
 
-    /* Start getting the data from the source. Request and transfer data
-     * from the source and display a modal progress dialog while doing this. */
+    /* Indicate to the guest that we have dropped the data on the host.
+     * The guest then will initiate the actual "drop" operation into our proxy on the guest. */
     Assert(!m_dndSource.isNull());
-    CProgress progress = m_dndSource.Drop(strMimeType,
+    CProgress progress = m_dndSource.Drop(strMIMEType,
                                           UIDnDHandler::toVBoxDnDAction(dropAction));
+    LogFlowFunc(("Source: isOk=%RTbool\n", m_dndSource.isOk()));
     if (m_dndSource.isOk())
     {
+        /* Send a mouse event with released mouse buttons into the guest that triggers
+         * the "drop" event in our proxy window on the guest. */
+        AssertPtr(m_pSession);
+        m_pSession->mouse().PutMouseEvent(0, 0, 0, 0, 0);
+
         msgCenter().showModalProgressDialog(progress,
                                             tr("Retrieving data ..."), ":/progress_dnd_gh_90px.png",
                                             m_pParent);
 
-        LogFlowFunc(("fCanceled=%RTbool, fCompleted=%RTbool, isOk=%RTbool, hrc=%Rhrc\n",
+        LogFlowFunc(("Progress: fCanceled=%RTbool, fCompleted=%RTbool, isOk=%RTbool, hrc=%Rhrc\n",
                      progress.GetCanceled(), progress.GetCompleted(), progress.isOk(), progress.GetResultCode()));
 
         if (!progress.GetCanceled())
@@ -546,44 +683,17 @@ int UIDnDHandler::retrieveDataInternal(      Qt::DropAction  dropAction,
             if (RT_SUCCESS(rc))
             {
                 /* After we successfully retrieved data from the source we query it from Main. */
-                QVector<uint8_t> vecData = m_dndSource.ReceiveData();
-                if (!vecData.isEmpty())
+                vecData = m_dndSource.ReceiveData(); /** @todo QVector.size() is "int" only!? */
+                if (m_dndSource.isOk())
                 {
-                    switch (vaType)
-                    {
-                        case QVariant::String:
-                        {
-                            vaData = QVariant::fromValue(QString(reinterpret_cast<const char *>(vecData.constData())));
-                            Assert(vaData.type() == QVariant::String);
-                            break;
-                        }
-
-                        case QVariant::ByteArray:
-                        {
-                            QByteArray ba(reinterpret_cast<const char*>(vecData.constData()), vecData.size());
-
-                            vaData = QVariant::fromValue(ba);
-                            Assert(vaData.type() == QVariant::ByteArray);
-                            break;
-                        }
-
-                        case QVariant::StringList:
-                        {
-                            QString strData = QString(reinterpret_cast<const char*>(vecData.constData()));
-                            QStringList lstString = strData.split("\r\n", QString::SkipEmptyParts);
-
-                            vaData = QVariant::fromValue(lstString);
-                            Assert(vaData.type() == QVariant::StringList);
-                            break;
-                        }
-
-                        default:
-                            rc = VERR_NOT_SUPPORTED;
-                            break;
-                    }
+                    if (vecData.isEmpty())
+                        rc = VERR_NO_DATA;
                 }
                 else
-                    rc = VERR_NO_DATA;
+                {
+                    msgCenter().cannotDropDataToHost(m_dndSource, m_pParent);
+                    rc = VERR_GENERAL_FAILURE; /** @todo Fudge; do a GetResultCode() to rc translation. */
+                }
             }
             else
                 msgCenter().cannotDropDataToHost(progress, m_pParent);
@@ -597,24 +707,25 @@ int UIDnDHandler::retrieveDataInternal(      Qt::DropAction  dropAction,
         rc = VERR_GENERAL_FAILURE; /** @todo Fudge; do a GetResultCode() to rc translation. */
     }
 
-    setMode(DNDMODE_UNKNOWN);
+    setOpMode(DNDMODE_UNKNOWN);
 
     LogFlowFuncLeaveRC(rc);
     return rc;
 }
 
-void UIDnDHandler::setMode(DNDMODE enmMode)
+void UIDnDHandler::setOpMode(DNDOPMODE enmMode)
 {
     QMutexLocker AutoWriteLock(&m_WriteLock);
-    m_enmMode = enmMode;
-    LogFlowFunc(("Mode is now: %RU32\n", m_enmMode));
+    m_enmOpMode = enmMode;
+    LogFunc(("Operation mode is now: %RU32\n", m_enmOpMode));
 }
 
-int UIDnDHandler::sltGetData(const QString        &strMimeType,
+int UIDnDHandler::sltGetData(      Qt::DropAction  dropAction,
+                             const QString        &strMIMEType,
                                    QVariant::Type  vaType,
                                    QVariant       &vaData)
 {
-    int rc = retrieveDataInternal(Qt::CopyAction, strMimeType, vaType, vaData);
+    int rc = retrieveData(dropAction, strMIMEType, vaType, vaData);
     LogFlowFuncLeaveRC(rc);
     return rc;
 }
@@ -695,4 +806,3 @@ Qt::DropActions UIDnDHandler::toQtDnDActions(const QVector<KDnDAction> &vecActio
     LogFlowFunc(("dropActions=0x%x\n", int(dropActs)));
     return dropActs;
 }
-
